print("import: 'pysmi'")
import pysmi

print("import: 'pysmi.borrower'")
import pysmi.borrower

print("import: 'pysmi.codegen'")
import pysmi.codegen

print("import: 'pysmi.lexer'")
import pysmi.lexer

print("import: 'pysmi.parser'")
import pysmi.parser

print("import: 'pysmi.reader'")
import pysmi.reader

print("import: 'pysmi.searcher'")
import pysmi.searcher

print("import: 'pysmi.writer'")
import pysmi.writer

